/*
 * Copyright (c) 2002 Tomas Svensson <tsn@gbdev.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "conf.h"

#include <errno.h>
#include <memory.h>
#include <stdio.h>
#include <unistd.h>

#ifndef HAVE_STRLCPY
#include "strlcpy.h"
#endif

#include "tlswrap.h"
#include "misc.h"

extern int debug;

void sys_err(const char *err){
	perror(err);
  	exit(1);
}

void user_close(struct user_data *ud)
{
	if (ud->connected == CONN_YES || ud->connected == CONN_IN_PROG)
		close(ud->serv_fd);
	close(ud->user_fd);

	if (ud->ssl_ctrl)
		SSL_free(ud->ssl_ctrl);

	memset(ud, 0, sizeof(struct user_data));
	ud->user_fd = -1;
	ud->serv_fd = -1;
	ud->connected = CONN_NO;
	ud->ssl_data_fd_mode = TLS_NONE;
	if (debug)
		printf("user_close\n");
	if (ud->data_connected != CONN_NO)
		data_close(ud);
}

void data_close(struct user_data *ud)
{
	if (ud->ssl_data) {
		SSL_free(ud->ssl_data);
		ud->ssl_data = NULL;
	}

	if (ud->data_connected == CONN_DATA_OK ||
	    ud->data_connected == CONN_DATA_LISTEN) 
		close(ud->user_data_fd);
	if (ud->data_connected == CONN_IN_PROG ||
	    ud->data_connected == CONN_DATA_TLS ||
	    ud->data_connected == CONN_DATA_OK) 
		close(ud->serv_data_fd);

	ud->user_data_fd = -1;
	ud->serv_data_fd = -1;

	ud->data_connected = CONN_NO;
	ud->ssl_data_fd_mode = TLS_NONE;
	if (debug)
		printf("data_close\n");
}

size_t
extr_str(const char *src, size_t src_len, char *dst, size_t dst_len) {

	/* Extract a \r\n (or just \n) terminated string from a binary
     	buffer. The copied string will always be null-terminated and
     	without \r\n. The function will return a pointer to the next
     	byte after the extracted string or NULL if no string could be
     	found */

  	char *ptr;
  	int ext_size,size;

  	if ((ptr = memchr(src,'\n',src_len)) == NULL)
		return 0;

  	ext_size = size = ptr - src;

  	if (src[ext_size-1] == '\r')
    		ext_size--;

  	if ( (ext_size + 1) <= dst_len) {
    		memcpy(dst, src, ext_size);
    		dst[ext_size] = 0;
  	} else {
    		printf("ext_str: dst buffer too small for extracted string\n");
    		return 0;
  	}

  	return size + 1;
}
int print_to_ud(struct user_data *ud, const char *s) {

	/* all pointers must be setup correctly prior to calling this
     	function, or it will segfault */

  	size_t slen;
  	char str[130];
  
  	snprintf(str, sizeof(str), s);

  	slen = strlen(str); /* NOT including null char */
  
  	if ( (&ud->s2u_buf[S2U_SIZE] - ud->s2u_i) < slen) {
    		printf("print_to_ud: can't fit string to buffer\n");
    		return 1;
  	} else {
    		memcpy(ud->s2u_i,str,slen);
    		ud->s2u_i+=slen;
  	}

 	return 0;
}

int print_to_serv(struct user_data *ud, const char *s) {

	size_t slen;
	char str[130];

	snprintf(str, sizeof(str), s);
	slen=strlen(str); /* NOT including null char */
	if ( (&ud->u2s_buf[U2S_SIZE]-ud->u2s_i)<slen) {
		printf("print_to_ud: can't fit string to buffer\n");
		return 1; 
	} else {
		memcpy(ud->u2s_i,str,slen);
		ud->u2s_i+=slen;
	}
	
	return 0;
}
#if 1 
int find_max_fd(int listen, struct user_data *ud, int max_users) {
	int i, max_fd=listen;
 
  	for(i = 0; i < max_users; i++) {
    		if (ud[i].user_fd != -1) {
      			if (ud[i].user_fd > max_fd)
				max_fd = ud[i].user_fd;
      			if ((ud[i].serv_fd > max_fd) &&
			    ((ud[i].connected == CONN_YES)||
	  	   	    (ud[i].connected == CONN_IN_PROG )))
				max_fd = ud[i].serv_fd;
      			if ( (ud[i].serv_data_fd > max_fd) &&
			    ((ud[i].data_connected == CONN_YES) ||
	  	    	    (ud[i].data_connected == CONN_IN_PROG )))
				max_fd = ud[i].serv_data_fd; 
      			if (ud[i].data_connected == CONN_DATA_LISTEN)
				if (ud[i].user_data_fd > max_fd)
					max_fd = ud[i].user_data_fd;
      			if (ud[i].ssl_data_fd_mode != TLS_NONE)
	 			if (ud[i].serv_data_fd > max_fd)
					max_fd = ud[i].serv_data_fd;
      			if (ud[i].data_connected == CONN_DATA_OK) {
				if (ud[i].user_data_fd > max_fd)
					max_fd = ud[i].user_data_fd;
				if (ud[i].serv_data_fd > max_fd)
					max_fd = ud[i].serv_data_fd;
			}
      		}
    	}
  
  	return max_fd;
}
#else

int find_max_fd(fd_set *fd_r, fd_set *fd_w)
{
	int i, max_fd;

	max_fd = 0;

	for (i = FD_SETSIZE - 1; i >= 0 ; i--) {
		if ( FD_ISSET(i,fd_r) || FD_ISSET(i,fd_w) )
			return i;
	}

	return 0;
}
#endif
void init_ud(struct user_data *ud, int max_users) {

	/* Initializes the user_data structures */

	int i;
  	memset(&ud[0], 0, max_users * sizeof(struct user_data));
  	for (i = 0; i < max_users; i++)
    		ud[i].user_fd = -1;
}

int find_free_slot(struct user_data *ud, int max_users) {

	/* Returns the index of the first free ud slot, otherwise
     	   it returns -1 */

  	int i;
  
  	for (i = 0; i < max_users; i++)
    		if (ud++->user_fd == -1) return i;
  
  	return -1;
}
