/*
 * Copyright (c) 2002 Tomas Svensson <tsn@gbdev.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "conf.h"

#define _POSIX_PII_SOCKET /* for Tru64 UNIX 5.1 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef HAVE_STRLCPY
#include "strlcpy.h"
#endif

extern int debug;

#include "tlswrap.h"
#include "network.h"
#include "misc.h"

void setup_connect_1(struct user_data *ud, int index, char *serv,
		     char *port, int write_pipe) {
	struct dns_msg dns;

  	dns.ud = index;
  	strlcpy(dns.port, port, sizeof(dns.port));
  	strlcpy(dns.hostname, serv, sizeof(dns.hostname));
  	if (write(write_pipe, &dns, sizeof(dns)) != sizeof(dns)) {
    		fprintf(stderr, "Error: Too many hostname lookups\n");
    		return;
  	}
  	ud->connected = CONN_DNS;
  	if (debug)
		printf("Resolving %s...\n",serv);
}

void
setup_connect_2(struct user_data *ud, struct dns_msg *dns, int data)
{
	char *ep;
  	int result;

  	if (strlen(dns->hostname) == 0) {
    		if (debug)
			printf("Error: Could not resolve hostname\n");
    		if (!(data))
			ud->connected = CONN_CMD;
    		return;
	}
  	ud->rport = strtol(dns->port, &ep, 10);
	if (debug)
		printf("Connecting to %s on port %s, please wait...\n",
		    dns->hostname, dns->port);
  	if (data)
		ud->serv_data_fd = setup_connect(dns->hostname, dns->port,
		    &ud->lport, &result);
  	else
		ud->serv_fd = setup_connect(dns->hostname, dns->port,
		    &ud->lport, &result);
  	if (result == 0) {
    		if (data) {
			if (debug)
				printf("data connected\n");
			ud->data_connected = CONN_YES;
    		} else {
    			ud->connected = CONN_YES;
    			ud->serv_status = SERV_CONN;
    		}
  	} else if (result == 2) {
    		print_to_ud(ud, "Can't find hostname, should not happen!");
    		ud->connected = CONN_CMD;
  	} else if (result == 3) {
		print_to_ud(ud,"421 Connection refused by server\r\n");
		user_close(ud);
  	} else {
		if (!data)
			ud->connected = CONN_IN_PROG;  
		else
			ud->data_connected = CONN_IN_PROG;
  	} 
}

int
setup_connect(const char *host, const char *port,
	unsigned int *lport, int *result)
{

	int conn_fd, flags,  sockopt;
	char *ep;
#if 0
	int error;
  	struct addrinfo hints, *res;
	char lports[6];
#else
	struct sockaddr_in sin, *sin2;
	unsigned short nport;
#endif
  	struct sockaddr sa;
  	size_t sa_len;
 
#if 0
  	memset(&hints,0,sizeof(hints));

	hints.ai_flags  = AI_NUMERICHOST; /* No dns lookup */
	hints.ai_family = pf; /* PF_INET or PF_INET6 */
  	hints.ai_socktype = SOCK_STREAM;
  
  	error = getaddrinfo(host, port, &hints, &res);
  	if (error) {
    		*result = 2;
    		return -1;
  	}

  	conn_fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
#else
	conn_fd = socket(PF_INET, SOCK_STREAM, 0);
#endif
  	if (conn_fd < 0)
		sys_err("setup_connect_socket");

  	flags = fcntl(conn_fd,F_GETFL);
  	fcntl(conn_fd, F_SETFL, flags | O_NONBLOCK);

  	sockopt = 1;
  	if (setsockopt(conn_fd, SOL_SOCKET, SO_KEEPALIVE, &sockopt, sizeof(sockopt)))
    		sys_err("setsockopt-keepalive");
/*
  	memset(&hints,0,sizeof(hints));
  
  	hints.ai_flags  = AI_NUMERICHOST | AI_PASSIVE; // No dns lookup 
  	hints.ai_family = pf;
  	hints.ai_socktype = SOCK_STREAM;

  if (vhost[0]!=0) {
    error = getaddrinfo(vhost,NULL, &hints, &res2);
    if (bind(conn_fd, res2->ai_addr, res2->ai_addrlen) < 0)
      perror("bind");
    freeaddrinfo(res2);
  } else {
    */
    	memset(&sa,0,sizeof(sa));
	sa.sa_family = PF_INET;
    	if (bind(conn_fd, &sa, sizeof(sa)) < 0)
      		perror("bind");
 /* }
  if (error) {
    printf("%s", gai_strerror(error));
    exit(0);
  } */

  	sa_len = sizeof(sa);
  	if (getsockname(conn_fd, &sa, &sa_len))
		sys_err("getsockname");

#if 0
  	if (getnameinfo(&sa,sizeof(sa),NULL,0,lports,sizeof(lports),NI_NUMERICSERV))
    		sys_err("getnameinfo");
 
  	*lport = strtol(lports, &ep, 10);
  	if (connect(conn_fd, res->ai_addr, res->ai_addrlen) != 0) {
#else
	memset(&sin, 0, sizeof(sin));
	sin2 = (struct sockaddr_in *)&sa;
#ifdef HAVE_INET_ATON
	if (inet_aton(host, &sin.sin_addr) != 1)
		sys_err("inet_aton");
#else
	if ( (sin.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE)
		sys_err("inet_addr");
#endif
	*lport = ntohs(sin2->sin_port);
	nport = (unsigned short)strtol(port, &ep, 10);
	sin.sin_port = htons(nport);
	sin.sin_family = PF_INET;
	if (connect(conn_fd, (struct sockaddr *)&sin, sizeof(sin)) != 0) {
#endif
    		if (errno == ECONNREFUSED) {
			*result = 3;
    		} else {
			if (errno!=EINPROGRESS)
				sys_err("connect");
    			*result = 1; /* Nonblocking operation */
    		}
  	} else
		*result = 0; /* Connected */
#if 0
  	freeaddrinfo(res);
#endif
  	return conn_fd;
}

void dns_helper(int read_fd, int write_fd) {
	struct dns_msg dns;
  	ssize_t bytes;
#if 0
  	struct addrinfo hints, *res;
	int error;
#else
	struct sockaddr_in saddr;
	struct hostent *hptr;
#endif
#ifdef HAVE_SETPROCTITLE
  	char sp[40];
  	unsigned int serv = 0;
#endif
  	for(;;) {
#ifdef HAVE_SETPROCTITLE
    		snprintf(sp, sizeof(sp), "tlswrap-dns (serviced %u reqs)",
		    serv++);
    		setproctitle(sp);
#endif
    		bytes = read(read_fd, &dns, sizeof(dns));
    		if (bytes == 0) {
      			printf("Parent died, exiting...\n");
      			exit(0);
    		}
#if 0
		memset(&hints, 0, sizeof(hints));
		hints.ai_family = pf;
		hints.ai_socktype = SOCK_STREAM;
    		error = getaddrinfo(dns.hostname, NULL, &hints, &res);
    		if (error)
      			dns.hostname[0] = '\0';
    		else {
      			if (getnameinfo(res->ai_addr, res->ai_addrlen, dns.hostname,sizeof(dns.hostname), NULL, 0,
			    NI_NUMERICHOST)) sys_err("getnameinfo");
      			freeaddrinfo(res);
    		}
#else
		/* Old-style code for LEGACY os's */
		memset(&saddr, 0, sizeof(saddr));
		if ((hptr = gethostbyname(dns.hostname)) == NULL)
			dns.hostname[0] = '\0';
		else    {
			memcpy(&saddr.sin_addr, hptr->h_addr,
			    sizeof(saddr.sin_addr));
			strlcpy(dns.hostname, inet_ntoa(saddr.sin_addr),
			    sizeof(dns.hostname));
		}
#endif
    		bytes = write(write_fd, &dns, sizeof(dns));
  	}
  
}
#if 0
int setup_listen(sa_family_t pf, int max_users, const char *host, char *port,
		int portlen)
{

  /* Returns a non-blocking listening socket on the specified port.
     Listen backlog is set to max_users. if portlen != 0 then save
     port */
  
	int flags,listen_socket, error, sockopt;
  	char	*port2;
  	size_t slen;
  	struct sockaddr *sa;
  	char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];
        struct addrinfo hints, *res;
	struct sockaddr_storage ss;

	sa = (struct sockaddr*)&ss;
	
	memset(&ss, 0, sizeof(ss));
	
	if (portlen == 0)
		port2 = port;
	else
		port2 = NULL;

	memset(&hints, 0, sizeof(hints));
  	hints.ai_flags = AI_PASSIVE;
 	hints.ai_family = pf;
  	hints.ai_socktype = SOCK_STREAM;
  
  	error = getaddrinfo(host, port2, &hints, &res);
  	if (error) {
    		printf("%s", gai_strerror(error));
    		exit(0);
  	}

  	listen_socket = socket(res->ai_family, res->ai_socktype,
			    res->ai_protocol);

  	if (listen_socket < 0)
		sys_err("socket_listen_ipv6");

  	sockopt = 1;
  	if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &sockopt,
	    sizeof(sockopt)))
  		sys_err("setsockopt-reuseaddr");

  	if (bind(listen_socket, res->ai_addr, res->ai_addrlen) < 0)
    		sys_err("bind");
  
  	if (listen(listen_socket, max_users))
		sys_err("listen");

  	flags = fcntl(listen_socket,F_GETFL);
  	fcntl(listen_socket, F_SETFL, flags | O_NONBLOCK);

  	freeaddrinfo(res);

  	if (portlen > 0) {
		slen = sizeof(ss);
		if (getsockname(listen_socket, sa, &slen) < 0)
			sys_err("getsockname");
		error = (getnameinfo(sa, slen , hbuf, sizeof(hbuf), sbuf,
		    sizeof(sbuf), NI_NUMERICHOST | NI_NUMERICSERV));
	 	if (error) {
			printf("host = %s, port = %s, portlen = %d\n", host,
			    port,portlen);
			printf("%s", gai_strerror(error));
			sys_err("getnameinfo");
		}
		
		strlcpy(port, sbuf, portlen);
  	}	

	return listen_socket;
}
#else
int setup_listen(int max_users, const char *host, char *port, int portlen)
{
	/* IP V4 only version */

	/* Returns a non-blocking listening socket on the specified port.
	 Listen backlog is set to max_users. if portlen != 0 then blah */
	   
	int flags,listen_socket, sockopt;
	unsigned short rport;
	size_t slen;

	struct sockaddr_in sin;
	char *ep;
	struct hostent *hptr;
			
	rport = (unsigned short)strtol(port, &ep, 10);

	memset(&sin, 0, sizeof(sin));
	sin.sin_family = PF_INET;
	/* sin.sin_addr.s_addr = htonl(INADDR_ANY); */ 
	sin.sin_port = htons(rport); 
	
	if (host == NULL)
		sin.sin_addr.s_addr = htonl(INADDR_ANY);
	else {
		if ((hptr = gethostbyname(host)) == NULL)
				sys_err("can't resolve specified local hostname");
			else
				memcpy(&sin.sin_addr, hptr->h_addr, sizeof(sin.sin_addr));
	}
	listen_socket = socket(PF_INET, SOCK_STREAM, 0);

        if (listen_socket < 0) 
		sys_err("socket_listen_ipv4");

	sockopt = 1;
	if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &sockopt,
	    sizeof(sockopt)))
		sys_err("setsockopt-reuseaddr");

	if (bind(listen_socket, (struct sockaddr *)&sin, sizeof(sin)))
		sys_err("bind");

	slen = sizeof(struct sockaddr_in);
        if (getsockname(listen_socket, (struct sockaddr*)&sin, &slen) < 0)
		sys_err("getsockname");

	if (listen(listen_socket, max_users))
		sys_err("listen");

        flags = fcntl(listen_socket, F_GETFL);
	fcntl(listen_socket, F_SETFL, flags | O_NONBLOCK);

	if (portlen > 0)
		snprintf(port, portlen, "%u", ntohs(sin.sin_port));

	return listen_socket;
}
#endif

int get_local_ip(int fd, char *ip, int iplen)
{
	size_t slen;
	struct sockaddr_in sin;

        slen = sizeof(struct sockaddr_in); 
	if (getsockname(fd, (struct sockaddr*)&sin, &slen) < 0)
		sys_err("getsockname");

	strlcpy(ip, inet_ntoa(sin.sin_addr), iplen);
	return 0;
}
