/*
 * Copyright (c) 2002 Tomas Svensson <tsn@gbdev.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <strings.h>
#include <unistd.h>

#include "parse.h"
#include "misc.h"
#include "network.h"
#include "tls.h"

extern int debug;

struct ftp_cmd {
	char *cmd;
	int dir;
};

void 
intercept_user_buf(struct user_data * ud, char *buf, int len)
{
	int i;

	struct ftp_cmd cmd[] = {
		{"LIST", DATA_DOWN},
		{"RETR", DATA_DOWN},
		{"NLST", DATA_DOWN},
		{"STOR", DATA_UP},
		{"APPE", DATA_UP},
		{(char *)NULL, 0}
	};


	if (len < 4)
		return;

	for (i = 0; cmd[i].cmd; i++)
		if (memcmp(buf, cmd[i].cmd, 4) == 0) {
			ud->data_direction = cmd[i].dir;
			if (debug)
				printf("set direction for %s\n", cmd[i].cmd);
		}
}

int 
change_serv_buf(int pf, struct user_data * ud, char *buf, int *len)
{
	unsigned int port;
	char ip[17];
	int r;
	struct dns_msg dns;

	if (ud->prot == 'C')
		return 0;

	if (memcmp(buf, "227 ", 4) == 0) {	/* PASV reply detected */
		if ((r = pasv_to_ipport(buf, *len, ip, sizeof(ip), &port))
		    == 0) {
			if (debug)
				printf("ip: %s port: %u\n", ip, port);
			strlcpy(ud->serv_data_host, ip, sizeof(ud->serv_data_host));
			snprintf(ud->serv_data_port, sizeof(ud->serv_data_port), "%u", port);
			strlcpy(dns.hostname, ip, sizeof(dns.hostname));
			strlcpy(dns.port, ud->serv_data_port, sizeof(dns.port));
			setup_connect_2(pf, ud, &dns, 1);
			if (ud->data_connected == CONN_YES) {
				open_local_dataport(pf, ud);
			}
			return 1;
		} else
			printf("change_serv_buf failed %d\n", r);
	}
	return 0;
}

void 
open_local_dataport(int pf, struct user_data * ud)
{
	char port[6];
	char pasv[25];
	char tmp[80], *ep;
	char myip[NI_MAXHOST];

	get_local_ip(ud->user_fd, myip, sizeof(myip));
	ud->user_data_fd = setup_listen(pf, 5, myip, port, sizeof(port));
	if (debug)
		printf("open_local_dataport: fd = %d, port = %s\n", ud->user_data_fd, port);
	ipport_to_pasv(pasv, sizeof(pasv), myip, strtol(port, &ep, 10));
	snprintf(tmp, sizeof(tmp), "227 Entering Passive Mode (%s)\r\n", pasv);
	if (debug)
		printf("sending %s", tmp);
	write(ud->user_fd, tmp, strlen(tmp));
	ud->data_connected = CONN_DATA_LISTEN;
}

int 
pasv_to_ipport(char *buf, int len, char *ip, int iplen, unsigned int *port)
{
	char *ep, *ptr, *ptr2;
	int i;
	int num;

	if ((ptr = strchr(buf, '(')) == NULL)
		return 1;

	ptr2 = ++ptr;
	for (i = 0; i < 4; i++) {
		if ((ptr = strchr(++ptr, ',')) == NULL)
			return 2;
		*ptr = '.';
	}
	*ptr++ = '\0';
	strlcpy(ip, ptr2, iplen);
	ptr2 = ptr;
	if ((ptr = strchr(ptr, ',')) == NULL)
		return 3;
	*ptr++ = '\0';
	num = strtol(ptr2, &ep, 10);
	if (num <= 0 || *ep != '\0')
		return 4;
	*port = num * 256;
	ptr2 = ptr;
	if ((ptr = strchr(ptr, ')')) == NULL)
		return 5;
	*ptr = '\0';
	num = strtol(ptr2, &ep, 10);
	if (num < 0 || *ep != '\0') {
		printf("pasv_to_ipport FAILED with %s that got changed into %d (at %s)\n", ptr2, num, ep);
		return 6;
	}
	*port += num;

	return 0;
}

void
ipport_to_pasv(char *buf, int len, const char *ip, unsigned int port)
{
	char *ptr;

	strlcpy(buf, ip, len);
	while ((ptr = strchr(buf, '.')))
		*ptr = ',';
	snprintf(buf, len, "%s,%d,%d", buf, port / 256, port % 256);
}

int
parse_serv_buf(struct user_data * ud, int index)
{
	int size;
	char dst[BUF_SIZE], s[100];

	if ((size = extr_str(ud->serv_input, BUF_SIZE, dst, sizeof(dst))) == 0)
		return 1;	/* Nothing could be extracted */


	if ((ud->serv_status == SERV_CONN) && (strncasecmp(dst, "220 ", 4) == 0)) {
		print_to_serv(ud, "AUTH TLS\r\n");
		ud->serv_status = SERV_AUTH;
	} else if ((ud->serv_status == SERV_AUTH) && (strncasecmp(dst, "234 ", 4) == 0)) {
		ud->serv_status = SERV_TLS;
		tls_auth(ud, 0);
	} else if ((ud->serv_status == SERV_TLS_OK) && (strncasecmp(dst, "200 ", 4) == 0)) {
		ud->serv_status = SERV_PBSZ;
		snprintf(s, sizeof(s), "PROT %c\r\n", ud->prot);
		if (debug)
			printf(s);
		print_to_serv(ud, s);
	} else if ((ud->serv_status == SERV_PBSZ) && (strncasecmp(dst, "200 ", 4) == 0)) {
		ud->serv_status = SERV_PROT;
		snprintf(s, sizeof(s), "USER %s\r\n", ud->user);
		print_to_serv(ud, s);
	} else if ((ud->serv_status == SERV_PROT) && (strncasecmp(dst, "331 ", 4) == 0)) {
		snprintf(s, sizeof(s), "PASS %s\r\n", ud->pass);
		print_to_serv(ud, s);
		ud->serv_status = SERV_FLOW;
	}
	memmove(ud->serv_input, &ud->serv_input[size], BUF_SIZE - size - 1);
	ud->serv_ptr -= size;
	return 0;
}

int
parse_buf(struct user_data * ud, int index, int dns_write_pipe)
{
	int size;
	char dst[BUF_SIZE], *ptr, s[100], tmp[200];

	if ((size = extr_str(ud->user_input, BUF_SIZE, dst, sizeof(dst))) == 0)
		return 1;	/* Nothing could be extracted */

	/* Parse starts here and is ugly */

	if ((ud->connected == CONN_NO) && (strncasecmp(dst, "USER ", 5) == 0)) {
		strlcpy(tmp, dst + 5, sizeof(tmp));
		ptr = memchr(tmp, '@', strlen(tmp));
		if (ptr != NULL) {
			*ptr++ = 0;
			if ((strlen(tmp) > 0) && (strlen(ptr) > 0)) {
				if (tmp[0] == '#') {
					ud->prot = 'C';	/* Encrypt control only */
					strlcpy(ud->user, tmp + 1, sizeof(ud->user));
				} else {
					ud->prot = 'P';	/* Encrypt everything */
					strlcpy(ud->user, tmp, sizeof(ud->user));
				}
				strlcpy(ud->serv_host, ptr, sizeof(ud->serv_host));
				ptr = memchr(ud->serv_host, ':', strlen(ud->serv_host));
				if (ptr != NULL) {
					*ptr++ = 0;
					if (strlen(ptr) > 0)
						strlcpy(ud->serv_port, ptr, sizeof(ud->serv_port));
					else
						ptr = NULL;
				}
				if (ptr == NULL)
					strlcpy(ud->serv_port, "21", sizeof(ud->serv_port));
				ud->connected = CONN_USER;
				if (debug)
					printf("Username: %s, Host: %s, Port: %s\n", ud->user,
					    ud->serv_host, ud->serv_port);
				snprintf(s, sizeof(s), "331 Password required for %s.\r\n", ud->user);
				print_to_ud(ud, s);
			}
		} else {
			if (debug)
				printf("don't find any @\n");
			user_close(ud);
		}
	} else
		/* Attempt connection directly after receiving PASS */

	if ((ud->connected == CONN_USER) && (strncasecmp(dst, "PASS ", 5) == 0)) {
		strlcpy(ud->pass, dst + 5, sizeof(ud->pass));
		ud->connected = CONN_PASS;
		setup_connect_1(ud, index, ud->serv_host, ud->serv_port, dns_write_pipe);
	}
	memmove(ud->user_input, &ud->user_input[size], BUF_SIZE - size - 1);
	ud->user_ptr -= size;
	return 0;
}
