#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <arpa/ftp.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <pwd.h>
#include <ctype.h>
#include <openssl/ssl.h>
#include <openssl/rand.h>
#include <openssl/err.h>
#include <fcntl.h>

#include "tlswrap.h"
#include "tls.h"
#include "misc.h"

extern int debug;

SSL_CTX *tls_ctx;

void tls_init(char *egd_sock)
{
#ifdef HAVE_RAND_STATUS
	int egdbytes;
#endif

	SSL_load_error_strings();                /* readable error messages */
	SSL_library_init();                      /* initialize library */

	tls_ctx = SSL_CTX_new(SSLv23_method());
	
	if (!tls_ctx) {
		printf("SSL_CTX_new() %s\r\n",(char *)ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	if (debug)
		printf("egd_sock is %s\n", egd_sock);
#ifdef HAVE_RAND_STATUS
	if (RAND_status() != 1) {
		if ( (egdbytes = RAND_egd(egd_sock)) == -1 ) {
			fprintf(stderr, "egd_sock is %s\n", egd_sock);
			sys_err("RAND_egd failed\n");
		}
		if (RAND_status() != 1) {	
			printf("ssl_init: System without /dev/urandom, PRNG seeding must be done manually, but isn't\r\n");
			exit(1);
		}
	}
#endif
	SSL_CTX_set_options(tls_ctx, SSL_OP_NO_SSLv2);
	SSL_CTX_set_default_verify_paths(tls_ctx);

	SSL_CTX_set_session_cache_mode(tls_ctx, SSL_SESS_CACHE_CLIENT);
	SSL_CTX_set_session_id_context(tls_ctx, (const unsigned char *) "1", 1);

	if (debug)
		printf("TLS initialization successful.\n");
}

void
tls_auth(struct user_data *ud, int data)
{
	SSL *ssl;

	ssl = SSL_new(tls_ctx);

	if (ssl == NULL) {
			printf("SSL_new() %s\r\n",(char *)ERR_error_string(ERR_get_error(), NULL));
			exit(1);
		}
	
	if (data)
		ud->ssl_data = ssl;
	else
		ud->ssl_ctrl = ssl;
	if (debug)
		printf("tls_auth: ciphers: %s\n", cfg_tlsciphers);

        SSL_set_cipher_list(ssl, cfg_tlsciphers);
	if (data) {
		if (SSL_set_fd(ssl, ud->serv_data_fd) != 1) printf("SSL_set_fd_error");
		tls_auth_cont(ud, 1);
	} else {
		if (SSL_set_fd(ssl, ud->serv_fd) != 1) printf("SSL_set_fd_error");
		tls_auth_cont(ud, 0);
	}
}

void
tls_auth_cont(struct user_data *ud, int data)
{
        int status, sslerr;
	SSL_CIPHER *cipher;
	char cipher_info[128];
	SSL *ssl;

	if (debug)
		printf("tls_auth_cont\n");
	ssl = (data) ? ud->ssl_data : ud->ssl_ctrl;
	
	if (ssl == NULL) printf("SSL == NULL!\n");

	status = SSL_connect(ssl);
	sslerr = SSL_get_error(ssl, status);
	if (data)
		ud->ssl_data_fd_mode = TLS_NONE;
	else
		ud->ssl_ctrl_fd_mode = TLS_NONE;

	if (status == 1) {
		cipher = SSL_get_current_cipher(ssl);
		SSL_CIPHER_description(cipher, cipher_info, sizeof(cipher_info));
		if (debug)
			printf("cipher %s\n", cipher_info);
		if (data) {
			ud->tls_status |= TLS_DATA;
			ud->data_connected = CONN_DATA_OK;
		} else {
			ud->serv_status = SERV_TLS_OK;
			ud->tls_status |= TLS_CTRL;
			print_to_serv(ud, "PBSZ 0\r\n");
			if (debug)
				printf("printed pbsz\n");
		}
	} else {
		switch (sslerr) {
			case SSL_ERROR_WANT_READ:
				if (debug)
					printf("setting TLS_READ\n");
				if (data)
					ud->ssl_data_fd_mode = TLS_READ;
				else
					ud->ssl_ctrl_fd_mode = TLS_READ;
				break;
			case SSL_ERROR_WANT_WRITE:
				if (debug)
					printf("setting TLS_WRITE\n");
				if (data)
					ud->ssl_data_fd_mode = TLS_WRITE;
				else
					ud->ssl_ctrl_fd_mode = TLS_WRITE;
				break;
			default:
				if (debug)
					printf("tls_auth_cont failed (%d)\n", sslerr);
				if (sslerr) perror("tls_auth_cont");
				data_close(ud);

		}
	}
}

int
tls_write(struct user_data *ud, const void *buf, int num, int data)
{
	SSL 	*ssl;
	int	status, sslerr;

	ssl = (data) ? ud->ssl_data : ud->ssl_ctrl;

	status = SSL_write(ssl, buf, num);
	sslerr = SSL_get_error(ssl, status);

	if (status == -1) {
		if (data)
			ud->ssl_data_func = TLS_WRITE;
		else
			ud->ssl_ctrl_func = TLS_WRITE;
		switch (sslerr) {
			case SSL_ERROR_WANT_READ:
				if (data)
					ud->ssl_data_fd_mode = TLS_READ;
				else
					ud->ssl_ctrl_fd_mode = TLS_READ;
				break;
			case SSL_ERROR_WANT_WRITE:
		 		if (data)
					ud->ssl_data_fd_mode = TLS_WRITE;
				else
					ud->ssl_ctrl_fd_mode = TLS_WRITE;
				break;
			default:
				if (debug)
					printf("tls_write_error\n");
				return -1;
		}
	} else {
		if (data)
			ud->ssl_data_fd_mode = TLS_NONE;
		else
			ud->ssl_ctrl_fd_mode = TLS_NONE;
	}

	return status;
}

int
tls_read(struct user_data *ud, void *buf, int num, int data) 
{
	SSL     *ssl;
	int     status, sslerr;

	ssl = (data) ? ud->ssl_data : ud->ssl_ctrl;

	status = SSL_read(ssl, buf, num);
	sslerr = SSL_get_error(ssl, status);

	if (status == -1) {
		if (data)
			ud->ssl_data_func = TLS_READ;
		else
			ud->ssl_ctrl_func = TLS_READ;
		switch (sslerr) {
			case SSL_ERROR_WANT_READ:
				if (data)
					ud->ssl_data_fd_mode = TLS_READ;
				else
					ud->ssl_ctrl_fd_mode = TLS_READ;
				break;
			case SSL_ERROR_WANT_WRITE:
				if (data)
					ud->ssl_data_fd_mode = TLS_WRITE;
				else
					ud->ssl_ctrl_fd_mode = TLS_WRITE;
				break;
			default:
				if (debug)
					printf("tls_read_error\n");
				return -2;
		}
	} else {
		if (data)
			ud->ssl_data_fd_mode = TLS_NONE;
		else
			ud->ssl_ctrl_fd_mode = TLS_NONE;
	}
	
	return status;
}
