/*
 * Copyright (c) 2002 Tomas Svensson <tsn@gbdev.net>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define _POSIX_PII_SOCKET /* for Tru64 UNIX 5.1 */

#include "conf.h"

#include <sys/errno.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <netdb.h>

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "tlswrap.h"
#include "network.h"
#include "misc.h"
#include "parse.h"
#include "tls.h"
#include "config.h"

char *cfg_tlsrsafile;
char *cfg_tlsciphers;

int debug;

int
main(int argc, char *argv[])
{
	struct sigaction sact;
	int 	listen_fd;
	int	dns_write_pipe, dns_read_pipe, pipe1[2], pipe2[2];
	int	flags, idx;
	int	i, sel, newsock;
	int	remove_this,  serv_write;
	char	fakebuf[1];
	char 	cfg_listenport[6];
	char	cfg_egdsock[NI_MAXHOST];
	char	cfg_listenhost[NI_MAXHOST];
	unsigned int cfg_max_users;
	pid_t 	childpid;
	fd_set  rset, wset;
	struct sockaddr sockaddr;
	size_t socklen;
	struct	user_data *ud;
	struct	dns_msg dns;
	unsigned int	tcpbufsize, tcpsndlowat;
	ssize_t bytes, bytesW;
	struct rlimit rlimit;
	char	token[4];

	if ( (cfg_tlsciphers = malloc(1024)) == NULL)
		exit(1);
	
	rlimit.rlim_cur = RLIM_INFINITY;
	rlimit.rlim_max = RLIM_INFINITY;
	setrlimit(RLIMIT_CORE, &rlimit);

	read_config(argc, argv, &cfg_max_users, cfg_listenport, 
	    sizeof(cfg_listenport), &debug,
	    cfg_egdsock, sizeof(cfg_egdsock), cfg_tlsciphers,
	    1024, &tcpbufsize, &tcpsndlowat, cfg_listenhost,
	    sizeof(cfg_listenhost), token, sizeof(token));

	if ( (ud = (struct user_data*)malloc(cfg_max_users *
		    sizeof(struct user_data))) == NULL) {
		fprintf(stderr,"can't malloc user_data");
		exit(1);
	}

	tls_init(cfg_egdsock);

	sact.sa_handler = SIG_IGN;
	sigemptyset(&sact.sa_mask);
	sact.sa_flags = 0;
	sigaction(SIGPIPE, &sact, NULL);

#ifdef HAVE_SETPROCTITLE
  	setproctitle("tlswrap");
#endif

	if (pipe(pipe1)) sys_err("pipe1");
	if (pipe(pipe2)) sys_err("pipe2");

	if ( (childpid = fork()) == 0) {
		close(pipe1[1]);
		close(pipe2[0]);
		dns_helper(pipe1[0], pipe2[1]);
	} else {
		close(pipe1[0]);
		close(pipe2[1]);
	}

	dns_write_pipe = pipe1[1];
	dns_read_pipe  = pipe2[0];

	/* Do blocking DNS requests before this (if any) */

	flags = fcntl(dns_write_pipe, F_GETFL);
	fcntl(dns_write_pipe, F_SETFL, flags | O_NONBLOCK);
        flags = fcntl(dns_read_pipe, F_GETFL);
	fcntl(dns_read_pipe, F_SETFL, flags | O_NONBLOCK);

	init_ud(ud, cfg_max_users);

	listen_fd = setup_listen(5, cfg_listenhost, cfg_listenport, 0);

	fprintf(stderr,
	    "TLS-Wrap v0.7 (c) 2002 Tomas Svensson <tsn@gbdev.net>\n");
	fprintf(stderr,"Listening to clients on port %s\n", cfg_listenport);
#ifndef __CYGWIN__
#ifdef HAVE_DAEMON
	if (!debug)
		daemon(0 ,0);
#else
	if (!debug) {
		if ( (childpid = fork()) < 0)
			sys_err("fork()");
		else if (childpid != 0)
			exit(0); /* parent */

		setsid();
		chdir("/");
	}
#endif /* !HAVE_DAEMON */
#endif /* !__CYGWIN__ */
 	for(;;) {
		FD_ZERO(&rset);
		FD_ZERO(&wset);

		FD_SET(listen_fd, &rset);
		FD_SET(dns_read_pipe, &rset);

		for(i = 0; i < cfg_max_users ; i++) {
			if (ud[i].user_fd != -1) {
				if (ud[i].u2s_i < &ud[i].u2s_buf[U2S_SIZE])
					FD_SET(ud[i].user_fd, &rset);
				if ((ud[i].connected == CONN_YES) &&
				    (ud[i].s2u_i < &ud[i].s2u_buf[S2U_SIZE]))
					FD_SET(ud[i].serv_fd, &rset);
				else if (ud[i].connected == CONN_IN_PROG) {
					FD_SET(ud[i].serv_fd, &rset);
					FD_SET(ud[i].serv_fd, &wset);
				}
				if (ud[i].data_connected == CONN_IN_PROG) {
					FD_SET(ud[i].serv_data_fd, &rset);
					FD_SET(ud[i].serv_data_fd, &wset);
				}
				if (ud[i].data_connected == CONN_DATA_OK) {
					if (ud[i].dc2s_i < 
					    &ud[i].dc2s_buf[DBUF_SIZE])
					 	if (ud[i].user_data_close !=
						    CLOSE_READ)
							FD_SET(ud[i].user_data_fd, &rset);
					
					if (ud[i].ds2c_i <
					    &ud[i].ds2c_buf[DBUF_SIZE])
						if  (ud[i].serv_data_close !=
						    CLOSE_READ)
							FD_SET(ud[i].serv_data_fd, &rset);
					
					if (ud[i].dc2s_i != ud[i].dc2s_o)
						if (ud[i].serv_data_close !=
						    CLOSE_WRITE)
							FD_SET(ud[i].serv_data_fd, &wset);
					if (ud[i].ds2c_i != ud[i].ds2c_o)
						if (ud[i].user_data_close != CLOSE_WRITE) FD_SET(ud[i].user_data_fd, &wset);
				} else if (ud[i].data_connected == CONN_DATA_LISTEN) {
					FD_SET(ud[i].user_data_fd, &rset);
					if (debug)
						printf("setting fd %d for conn_data_listen\n",ud[i].user_data_fd);
				}
			
			if ((ud[i].connected == CONN_YES) && (ud[i].u2s_i != ud[i].u2s_o))
				FD_SET(ud[i].serv_fd, &wset);
			if (ud[i].s2u_i != ud[i].s2u_o)
				if (memchr(ud[i].s2u_o, '\n', ud[i].s2u_i-ud[i].s2u_o) != NULL) /* should be? */
					FD_SET(ud[i].user_fd, &wset);

			/* TLS negotiation crap */

				if (ud[i].ssl_ctrl_fd_mode == TLS_READ) {
					if (debug)
						printf("TLS_READ: fd = %d\n", ud[i].serv_fd);
					FD_SET(ud[i].serv_fd, &rset);
					FD_CLR(ud[i].serv_fd, &wset);
				} else if (ud[i].ssl_ctrl_fd_mode == TLS_WRITE) {
					FD_SET(ud[i].serv_fd, &wset);
					FD_CLR(ud[i].serv_fd, &rset);
				}

                                if (ud[i].ssl_data_fd_mode == TLS_READ) {
					if (debug)
						printf("setting serv_data_fd (%d) in rset for TLS_READ\n", ud[i].serv_data_fd);
                                        FD_SET(ud[i].serv_data_fd, &rset); 
                                        FD_CLR(ud[i].serv_data_fd, &wset);
                                } else if (ud[i].ssl_data_fd_mode == TLS_WRITE) {
                                        FD_SET(ud[i].serv_data_fd, &wset);
                                        FD_CLR(ud[i].serv_data_fd, &rset);
                                }
			} /* of if fd */

		}

		/*
		max_fd = find_max_fd((dns_read_pipe > listen_fd) ?
		    dns_read_pipe : listen_fd, ud, cfg_max_users);
		 max_fd = find_max_fd(&rset, &wset);

		if (debug)
			printf("max_fd = %d\n",max_fd);
		*/
		sel = select(FD_SETSIZE, &rset, &wset, NULL, NULL);
		if (sel == -1)
			sys_err("select");
		
		if (FD_ISSET(dns_read_pipe, &rset)) {
			if ((bytes = read(dns_read_pipe, &dns, sizeof(dns))
			    == sizeof(dns))) {
				if ((ud[dns.ud].user_fd != -1) && (ud[dns.ud].connected == CONN_DNS))
					setup_connect_2(&ud[dns.ud], &dns, 0);
			} else
				sys_err("child died");
		}

		if (FD_ISSET(listen_fd, &rset)) {
			memset(&sockaddr, 0, sizeof(sockaddr));
			socklen = sizeof(sockaddr);
			if ( (newsock = accept(listen_fd, &sockaddr, &socklen)) != -1) {
				idx = find_free_slot(ud, cfg_max_users);
				if (idx==-1) {
					write(newsock,"No more users allowed\r\n",23);
					close(newsock);
				} else {
					ud[idx].u2s_i = ud[idx].u2s_o = ud[idx].u2s_buf;
					ud[idx].s2u_i = ud[idx].s2u_o = ud[idx].s2u_buf;
					ud[idx].user_ptr = ud[idx].user_input;
					ud[idx].serv_ptr = ud[idx].serv_input;
					ud[idx].connected = CONN_NO;
					ud[idx].data_connected = CONN_NO;
					ud[idx].serv_status = SERV_NONE;
					ud[idx].serv_data_fd = -1;
					ud[idx].user_fd = newsock;
					ud[idx].tls_status = 0;
					ud[idx].ssl_data_fd_mode = TLS_NONE;
					if (debug)
						printf("connected to user\n");
					print_to_ud(&ud[idx], "220 TLSWRAP FTP Proxy Server (Version 0.7) ready.\r\n");
				}
			}
		}

		for(i = 0; i < cfg_max_users; i++) {
			remove_this = 0;
			serv_write = 0;
			if (ud[i].user_fd!=-1) {

				/* TLS section */

				if (ud[i].serv_status == SERV_TLS) {
					if (debug)
						printf("checking TLS status\n");
					if ( ((ud[i].ssl_ctrl_fd_mode == TLS_READ) && FD_ISSET(ud[i].serv_fd, &rset)) || 
						((ud[i].ssl_ctrl_fd_mode == TLS_WRITE) && FD_ISSET(ud[i].serv_fd, &wset)) ) {
						tls_auth_cont(&ud[i], 0);
						continue;
					}
				}
				if (ud[i].data_connected == CONN_DATA_TLS) {
					if (debug)
						printf("checking conn_data_tls\n");
					if ( ((ud[i].ssl_data_fd_mode == TLS_READ) && FD_ISSET(ud[i].serv_data_fd, &rset)) ||
						((ud[i].ssl_data_fd_mode == TLS_WRITE) && FD_ISSET(ud[i].serv_data_fd, &wset)) ) {
						tls_auth_cont(&ud[i], 1);
						continue;
					}
				}

				/* TLS fd swapping */

				if (ud[i].tls_status & TLS_CTRL) {
				if ( ud[i].ssl_ctrl_fd_mode == TLS_READ && FD_ISSET(ud[i].serv_fd, &rset) &&
				      ud[i].ssl_ctrl_func == TLS_WRITE) {
					FD_SET(ud[i].serv_fd, &wset);
					FD_CLR(ud[i].serv_fd, &rset);
				} else if ( ud[i].ssl_ctrl_fd_mode == TLS_WRITE && FD_ISSET(ud[i].serv_fd, &wset) &&
				     ud[i].ssl_ctrl_func == TLS_READ) {
					FD_SET(ud[i].serv_fd, &rset);
					FD_CLR(ud[i].serv_fd, &wset);
				}}

                                if (ud[i].tls_status & TLS_DATA) {
                                        if ( ud[i].ssl_data_fd_mode == TLS_READ && FD_ISSET(ud[i].serv_data_fd, &rset) &&
                                          ud[i].ssl_data_func == TLS_WRITE) {
                                                FD_SET(ud[i].serv_data_fd, &wset);
                                                FD_CLR(ud[i].serv_data_fd, &rset);
                                        } else if ( ud[i].ssl_data_fd_mode == TLS_WRITE && FD_ISSET(ud[i].serv_data_fd, &wset) &&
                                          ud[i].ssl_data_func == TLS_READ) { 
                                                FD_SET(ud[i].serv_data_fd, &rset);
                                                FD_CLR(ud[i].serv_data_fd, &wset);
                                        }       
                                }       

				/* Read Section */

				if (ud[i].connected == CONN_YES) if (FD_ISSET(ud[i].serv_fd, &rset)) {
					if (ud[i].tls_status & TLS_CTRL)
						bytes = tls_read(&ud[i], ud[i].s2u_i, &ud[i].s2u_buf[S2U_SIZE] - ud[i].s2u_i,0);
					else
						bytes = read(ud[i].serv_fd, ud[i].s2u_i, &ud[i].s2u_buf[S2U_SIZE] - ud[i].s2u_i);

					if (bytes < 0) {
						if (errno != EWOULDBLOCK) {
							perror("server_read");
							data_close(&ud[i]);
						}
						continue;
					} else if (bytes==0) {
						FD_CLR(ud[i].serv_fd,&wset);
						FD_CLR(ud[i].user_fd,&rset);
						user_close(&ud[i]);
						remove_this=1;
					} else {
						ud[i].s2u_i+=bytes;
						FD_SET(ud[i].user_fd, &wset); /* Try to write this data below */
					}
				}
				
				if (ud[i].data_connected == CONN_DATA_OK)  {
					if (FD_ISSET(ud[i].serv_data_fd, &rset)) {
						if (ud[i].tls_status & TLS_DATA)
							bytes = tls_read(&ud[i], ud[i].ds2c_i, 
							    &ud[i].ds2c_buf[DBUF_SIZE] - ud[i].ds2c_i, 1);
						else
							bytes = read(ud[i].serv_data_fd, ud[i].ds2c_i,
							    &ud[i].ds2c_buf[DBUF_SIZE] - ud[i].ds2c_i);
						if (bytes < 0) {
							if (errno != EWOULDBLOCK) {
								perror("server_data_read");
								data_close(&ud[i]);
							}
							continue;
						} else if (bytes==0) {
								ud[i].serv_data_close = CLOSE_READ;
								if (debug)
									printf("setting CLOSE_READ in serv_data_close\n");
								if (ud[i].serv_read_cnt > 0)
									ud[i].user_data_close = CLOSE_READ;
						} else {
							if (debug)
								printf("read %ld bytes from serv_data_fd\n", (long)bytes);
							ud[i].ds2c_i += bytes;
							ud[i].serv_read_cnt += bytes;
							FD_SET(ud[i].user_data_fd, &wset); /* Try to write this data below */ 
						}
					}

					if (FD_ISSET(ud[i].user_data_fd, &rset)) {
						bytes = read(ud[i].user_data_fd, ud[i].dc2s_i,
						    &ud[i].dc2s_buf[DBUF_SIZE] - ud[i].dc2s_i);
						if (bytes < 0) {
							if (errno != EWOULDBLOCK) {
								perror("server_data_read");
								data_close(&ud[i]);
							}
							continue;
						} else if (bytes==0) {
								ud[i].user_data_close = CLOSE_READ;
								if (ud[i].user_read_cnt > 0)
									ud[i].serv_data_close = CLOSE_READ;
								if (debug)
									printf("setting CLOSE_READ in user_data_close\n");
						} else {
							if (debug)
								printf("read %ld bytes from user_data_fd\n", (long)bytes);
							ud[i].dc2s_i += bytes;
							ud[i].user_read_cnt += bytes;
							FD_SET(ud[i].serv_data_fd, &wset); /* Try to write this data below */
						}
					}

				}
				if (ud[i].connected == CONN_YES && ud[i].data_connected == CONN_DATA_LISTEN) {
					if (debug)
						printf("conn_data_listen\n");
					if (FD_ISSET(ud[i].user_data_fd, &rset)) {
							if (debug)
								printf("trying to accept user data connection\n");
							if ( (newsock = accept(ud[i].user_data_fd, &sockaddr, &socklen)) != -1) {
								close(ud[i].user_data_fd);
								ud[i].user_data_fd = newsock;
								ud[i].data_connected = CONN_DATA_OK;
								ud[i].ssl_data_fd_mode = TLS_NONE;
								ud[i].dc2s_i = ud[i].dc2s_o = ud[i].dc2s_buf;
								ud[i].ds2c_i = ud[i].ds2c_o = ud[i].ds2c_buf;
								ud[i].serv_data_close = CLOSE_NONE;
								ud[i].user_data_close = CLOSE_NONE;
								ud[i].user_read_cnt = 0;
								ud[i].serv_read_cnt = 0;
/*	
								(void)setsockopt(ud[i].user_data_fd, SOL_SOCKET, SO_SNDLOWAT, 
								    &tcpsndlowat, sizeof(tcpsndlowat));
								(void)setsockopt(ud[i].serv_data_fd, SOL_SOCKET, SO_SNDLOWAT,
								    &tcpsndlowat, sizeof(tcpsndlowat));
								(void)setsockopt(ud[i].user_data_fd, SOL_SOCKET, SO_SNDBUF,
								    &tcpbufsize, sizeof(tcpbufsize));
								(void)setsockopt(ud[i].serv_data_fd, SOL_SOCKET, SO_SNDBUF,
								    &tcpbufsize, sizeof(tcpbufsize));
								(void)setsockopt(ud[i].user_data_fd, SOL_SOCKET, SO_RCVBUF,
								    &tcpbufsize, sizeof(tcpbufsize));
								(void)setsockopt(ud[i].serv_data_fd, SOL_SOCKET, SO_RCVBUF,
								    &tcpbufsize, sizeof(tcpbufsize));
*/

								ud[i].data_connected = CONN_DATA_TLS;
								if (debug)
									printf("accept'ed client data connection\n");
								tls_auth(&ud[i], 1);
							} else
								printf("accept failed\n");
					}
				}

				if (ud[i].user_fd != -1) if (FD_ISSET(ud[i].user_fd,&rset)) {
					if ((bytes = read(ud[i].user_fd, ud[i].u2s_i, &ud[i].u2s_buf[U2S_SIZE]
					- ud[i].u2s_i)) < 0) {
						if (errno != EWOULDBLOCK) {
							perror("user_read");
							user_close(&ud[i]);
						}
					} else if (bytes==0) {
						user_close(&ud[i]);
					} else {
						ud[i].u2s_i += bytes;
						serv_write=1;
					}
				}

				/* Write Section */

				if (ud[i].data_connected == CONN_DATA_OK)  {
					if (FD_ISSET(ud[i].serv_data_fd, &wset) && ((bytes = ud[i].dc2s_i - ud[i].dc2s_o) > 0)) {
						if (ud[i].tls_status & TLS_DATA)
							bytesW = tls_write(&ud[i], ud[i].dc2s_o, bytes, 1);
						else
							bytesW = write(ud[i].serv_data_fd, ud[i].dc2s_o, bytes);
						if (bytesW < 0) {
							if (errno == EPIPE) {
								ud[i].serv_data_close = CLOSE_WRITE;
							}
							else if (errno != EWOULDBLOCK) {
								perror("serv_data_fd_write");
								data_close(&ud[i]);
							}
							continue;
						} else {
							if (debug)
								printf("wrote %ld bytes to serv_data_fd (of %ld requested)\n", 
								    (long)bytesW, (long)bytes);
							ud[i].dc2s_o += bytesW;
							if (ud[i].dc2s_o == ud[i].dc2s_i)
								ud[i].dc2s_o = ud[i].dc2s_i = ud[i].dc2s_buf;
						}
					}
						

					if (FD_ISSET(ud[i].user_data_fd, &wset) && ((bytes = ud[i].ds2c_i - ud[i].ds2c_o) > 0)) {
						bytesW = write(ud[i].user_data_fd, ud[i].ds2c_o, bytes);
						if (bytesW < 0) {
							if (errno == EPIPE) {
								ud[i].user_data_close = CLOSE_WRITE;
							}
							if (errno != EWOULDBLOCK) {
								perror("user_data_fd_write");
								data_close(&ud[i]);
							}
							continue;
						} else {
							 if (debug)
								printf("wrote %ld bytes to user_data_fd (of %ld requested)\n",
								    (long)bytesW, (long)bytes);
							ud[i].ds2c_o += bytesW;
							if (ud[i].ds2c_o == ud[i].ds2c_i)
								ud[i].ds2c_o = ud[i].ds2c_i = ud[i].ds2c_buf;
						}
					}

					if ( ( (ud[i].data_direction == DATA_DOWN && ud[i].serv_data_close == CLOSE_READ) ||
					     (ud[i].data_direction == DATA_UP && ud[i].user_data_close == CLOSE_READ) || 
					     (ud[i].serv_data_close == CLOSE_READ && ud[i].user_data_close == CLOSE_READ) ) &&
					     (ud[i].ds2c_o == ud[i].ds2c_i && ud[i].dc2s_o == ud[i].dc2s_i) )  {
						data_close(&ud[i]);
						if (debug)
							printf("data connection totally closed\n");
					}


					


				}

				if ((ud[i].connected == CONN_YES) || (ud[i].connected == CONN_IN_PROG)) {
					if (FD_ISSET(ud[i].serv_fd, &wset) && ((bytes = ud[i].u2s_i - ud[i].u2s_o) > 0))
						serv_write=1;
				} else if ((bytes = ud[i].u2s_i - ud[i].u2s_o) > 0) 
					serv_write=1;
				
				if (serv_write) {
					if (ud[i].connected != CONN_YES) {
						if ( (&ud[i].user_input[BUF_SIZE] - ud[i].user_ptr) >= bytes) {
							/* There is room in the user buffer for this data */
							memcpy(ud[i].user_ptr, ud[i].u2s_o, bytes);
							ud[i].u2s_o += bytes;
							ud[i].user_ptr += bytes;
							if (ud[i].u2s_o==ud[i].u2s_i)
								ud[i].u2s_o = ud[i].u2s_i = ud[i].u2s_buf;
						} else
							printf("could not copy user input to user buffer\n");
					} else {
						if (ud[i].prot == 'P')
							intercept_user_buf(&ud[i], ud[i].u2s_o, bytes);
						if (ud[i].tls_status & TLS_CTRL) 
							bytesW = tls_write(&ud[i], ud[i].u2s_o, bytes, 0);
						else
							bytesW = write(ud[i].serv_fd, ud[i].u2s_o, bytes);

						if (bytesW < 0) {
							if (errno != EWOULDBLOCK) {
								perror("server_write");
								user_close(&ud[i]);
							}
							continue;
						} else {
							ud[i].u2s_o += bytesW;
							if (ud[i].u2s_o == ud[i].u2s_i)
								ud[i].u2s_o = ud[i].u2s_i = ud[i].u2s_buf;
						}
					}
				}
			
				if ((bytes = ud[i].s2u_i - ud[i].s2u_o) > 0) {
					if ((ud[i].serv_status == SERV_FLOW || ud[i].connected != CONN_YES ) && FD_ISSET(ud[i].user_fd,&wset) && (memchr(ud[i].s2u_o, '\n', bytes) != NULL)) {
						if (change_serv_buf(&ud[i], ud[i].s2u_o, &bytes)) {
							bytesW = bytes;
						} else 
							bytesW = write(ud[i].user_fd, ud[i].s2u_o, bytes);

						if (bytesW < 0) {
							if (errno != EWOULDBLOCK) {
								perror("user_write");
								user_close(&ud[i]);
							}
						} else {
							ud[i].s2u_o+=bytesW;
							if (ud[i].s2u_o==ud[i].s2u_i)
								ud[i].s2u_o=ud[i].s2u_i=ud[i].s2u_buf;
						}
					}  else if ( (ud[i].serv_status != SERV_FLOW) && (ud[i].connected==CONN_YES)) {
						if ( (&ud[i].serv_input[BUF_SIZE] - ud[i].serv_ptr) >= bytes) {
							/* There is room in the server buffer for this data */
							if (debug)
								printf("eating server bytes\n");
							memcpy(ud[i].serv_ptr, ud[i].s2u_o, bytes);
							ud[i].s2u_o += bytes;
							ud[i].serv_ptr += bytes;
							if (ud[i].s2u_o==ud[i].s2u_i)
								ud[i].s2u_o = ud[i].s2u_i = ud[i].s2u_buf;
						} else
							printf("could not copy server input to server buffer\n");
					} 
				}
				

				/* Nonblocking connect to a remote data port gave a result */

				if (ud[i].data_connected == CONN_IN_PROG && ud[i].connected == CONN_YES) { 
					if (debug)
						printf("checking if %d is connected\n", ud[i].serv_data_fd);
					if (FD_ISSET(ud[i].serv_data_fd, &rset) ||
					    FD_ISSET(ud[i].serv_data_fd, &wset)) {
						if (read(ud[i].serv_data_fd, fakebuf, 0) < 0) {
							print_to_ud(&ud[i],"421 Connection failed");
							data_close(&ud[i]);
							if (debug)
								printf("data port connection failed\n");
						} else {
							ud[i].data_connected = CONN_YES;
							if (debug)
								printf("data port connected\n");
							open_local_dataport(&ud[i]);
						}
					}
				} 
			
				/* Nonblocking connect to remote server gave a result */

				if (ud[i].connected == CONN_IN_PROG) {
					if (FD_ISSET(ud[i].serv_fd,&rset) || FD_ISSET(ud[i].serv_fd,&wset)) {
						if (ud[i].connected == CONN_IN_PROG) {
							if (read(ud[i].serv_fd, fakebuf, 0) < 0) {
								print_to_ud(&ud[i],"421 Connection failed");
								user_close(&ud[i]);
							} else {
								ud[i].serv_status = SERV_CONN;
								ud[i].connected = CONN_YES;
								if (debug)
									printf("connected to server\n");
							}
						}
					}
				}
			}
			if (remove_this==1)
	      			ud[i].user_fd=-1;
			
			else	{
				if (ud[i].user_input != ud[i].user_ptr) 
					while (parse_buf(&ud[i], i, dns_write_pipe, token)==0);
				if (ud[i].serv_input != ud[i].serv_ptr)
					while (parse_serv_buf(&ud[i], i) == 0);
			}
		}
	}
}

